#ifndef __SEM_NAMES_h
#define __SEM_NAMES_h
static const char *semNames[] = {
  "no",
  "skip",
  "generic",
  "cdata",
  "htmllink",
  "htmltarget",
  "changetable",
  "noteref",
  "reverse",
  "configfile",
  "configstring",
  "configtweak",
  "document",
  "para",
  "heading1",
  "heading2",
  "heading3",
  "heading4",
  "heading5",
  "heading6",
  "heading7",
  "heading8",
  "heading9",
  "heading10",
  "contentsheader",
  "contents1",
  "contents2",
  "contents3",
  "contents4",
  "contents5",
  "contents6",
  "contents7",
  "contents8",
  "contents9",
  "contents10",
  "note",
  "notesheader",
  "notesdescription",
  "notranslate",
  "compbrl",
  "uncontracted",
  "contracted",
  "pagenum",
  "genpurp1",
  "genpurp2",
  "pagebreak",
  "attrtotext",
  "runninghead",
  "footer",
  "italicx",
  "boldx",
  "underlinex",
  "linespacing",
  "blankline",
  "softreturn",
  "newpage",
  "righthandpage",
  "code",
  "music",
  "math",
  "mi",
  "mn",
  "mo",
  "mtext",
  "mspace",
  "ms",
  "mglyph",
  "mrow",
  "mfrac",
  "msqrt",
  "mroot",
  "mstyle",
  "merror",
  "mpadded",
  "mphantom",
  "mfenced",
  "menclose",
  "msub",
  "msup",
  "msubsup",
  "munder",
  "mover",
  "munderover",
  "mmultiscripts",
  "none",
  "semantics",
  "mprescripts",
  "mtable",
  "mtr",
  "mtd",
  "maligngroup",
  "malignmark",
  "mlabeledtr",
  "maction",
  "chemistry",
  "graphic",
  "markhead",
  "utdmeta",
  "utdbrl",
  "utdnewpage",
  "utdnewline",
  "utdbrlonly",
  "utddispimg",
  "utdinlnimg",
  "end_all",
NULL
};
#endif /*__SEM_NAMES_H*/
