package org.liblouis;

public final class LogLevel
{
  public final static int ALL = -2147483648;
  public final static int DEBUG = 10000;
  public final static int INFO = 20000;
  public final static int WARNING = 30000;
  public final static int ERROR = 40000;
  public final static int FATAL = 50000;
  public final static int OFF = 2147483647;
}
