/* DO NOT EDIT THIS FILE - it is machine generated */
#include <jni.h>
/* Header for class org_liblouis_LibLouisUTDML */

#ifndef _Included_org_liblouis_LibLouisUTDML
#define _Included_org_liblouis_LibLouisUTDML
#ifdef __cplusplus
extern "C" {
#endif
#undef org_liblouis_LibLouisUTDML_NoContractions
#define org_liblouis_LibLouisUTDML_NoContractions 1L
#undef org_liblouis_LibLouisUTDML_CompbrlAtCursor
#define org_liblouis_LibLouisUTDML_CompbrlAtCursor 2L
#undef org_liblouis_LibLouisUTDML_DotsIO
#define org_liblouis_LibLouisUTDML_DotsIO 4L
#undef org_liblouis_LibLouisUTDML_Comp8Dots
#define org_liblouis_LibLouisUTDML_Comp8Dots 8L
#undef org_liblouis_LibLouisUTDML_Pass1Only
#define org_liblouis_LibLouisUTDML_Pass1Only 16L
#undef org_liblouis_LibLouisUTDML_CompbrlLeftCursor
#define org_liblouis_LibLouisUTDML_CompbrlLeftCursor 32L
#undef org_liblouis_LibLouisUTDML_OtherTrans
#define org_liblouis_LibLouisUTDML_OtherTrans 64L
#undef org_liblouis_LibLouisUTDML_UcBrl
#define org_liblouis_LibLouisUTDML_UcBrl 128L
#undef org_liblouis_LibLouisUTDML_DontInit
#define org_liblouis_LibLouisUTDML_DontInit 1073741824L
#undef org_liblouis_LibLouisUTDML_HtmlDoc
#define org_liblouis_LibLouisUTDML_HtmlDoc 536870912L
#undef org_liblouis_LibLouisUTDML_notUC
#define org_liblouis_LibLouisUTDML_notUC 268435456L
/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    initialize
 * Signature: (Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouisUTDML_initialize
  (JNIEnv *, jclass, jstring, jstring, jstring);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    version
 * Signature: ()Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_org_liblouis_LibLouisUTDML_version
  (JNIEnv *, jobject);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    loadXMLCatalog
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouisUTDML_loadXMLCatalog
  (JNIEnv *, jobject, jstring);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    translateString
 * Signature: (Ljava/lang/String;[B[B[ILjava/lang/String;Ljava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouisUTDML_translateString
  (JNIEnv *, jobject, jstring, jbyteArray, jbyteArray, jintArray, jstring, jstring, jint);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    backTranslateString
 * Signature: (Ljava/lang/String;[B[B[ILjava/lang/String;Ljava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouisUTDML_backTranslateString
  (JNIEnv *, jobject, jstring, jbyteArray, jbyteArray, jintArray, jstring, jstring, jint);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    translateFile
 * Signature: (Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouisUTDML_translateFile
  (JNIEnv *, jobject, jstring, jstring, jstring, jstring, jstring, jint);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    translateTextFile
 * Signature: (Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouisUTDML_translateTextFile
  (JNIEnv *, jobject, jstring, jstring, jstring, jstring, jstring, jint);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    backTranslateFile
 * Signature: (Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouisUTDML_backTranslateFile
  (JNIEnv *, jobject, jstring, jstring, jstring, jstring, jstring, jint);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    charToDots
 * Signature: (Ljava/lang/String;[B[BLjava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouisUTDML_charToDots
  (JNIEnv *, jobject, jstring, jbyteArray, jbyteArray, jstring, jint);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    dotsToChar
 * Signature: (Ljava/lang/String;[B[BLjava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouisUTDML_dotsToChar
  (JNIEnv *, jobject, jstring, jbyteArray, jbyteArray, jstring, jint);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    checkTable
 * Signature: (Ljava/lang/String;Ljava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouisUTDML_checkTable
  (JNIEnv *, jobject, jstring, jstring, jint);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    file2brl
 * Signature: ([Ljava/lang/String;)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouisUTDML_file2brl
  (JNIEnv *, jobject, jobjectArray);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    setWriteablePath
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouisUTDML_setWriteablePath
  (JNIEnv *, jobject, jstring);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    free
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouisUTDML_free
  (JNIEnv *, jobject);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    setLogFile
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouisUTDML_setLogFile
  (JNIEnv *, jobject, jstring);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    logEnd
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouisUTDML_logEnd
  (JNIEnv *, jobject);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    registerLogCallback
 * Signature: (Lorg/liblouis/LogCallback;)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouisUTDML_registerLogCallback
  (JNIEnv *, jobject, jobject);

/*
 * Class:     org_liblouis_LibLouisUTDML
 * Method:    setLogLevel
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouisUTDML_setLogLevel
  (JNIEnv *, jobject, jint);

#ifdef __cplusplus
}
#endif
#endif
/* Header for class org_liblouis_LibLouis */

#ifndef _Included_org_liblouis_LibLouis
#define _Included_org_liblouis_LibLouis
#ifdef __cplusplus
extern "C" {
#endif
/*
 * Class:     org_liblouis_LibLouis
 * Method:    charSize
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_org_liblouis_LibLouis_charSize
  (JNIEnv *, jobject);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    setDataPath
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouis_setDataPath
  (JNIEnv *, jobject, jstring);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    setLogFile
 * Signature: (Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouis_setLogFile
  (JNIEnv *, jobject, jstring);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    logEnd
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouis_logEnd
  (JNIEnv *, jobject);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    registerLogCallback
 * Signature: (Lorg/liblouis/LogCallback;)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouis_registerLogCallback
  (JNIEnv *, jobject, jobject);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    setLogLevel
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_org_liblouis_LibLouis_setLogLevel
  (JNIEnv *, jobject, jint);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    compileString
 * Signature: (Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouis_compileString
  (JNIEnv *, jobject, jstring, jstring, jstring);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    translateString
 * Signature: (Ljava/lang/String;[B[I[B[I[BLjava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouis_translateString
  (JNIEnv *, jobject, jstring, jbyteArray, jintArray, jbyteArray, jintArray, jbyteArray, jstring, jint);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    translate
 * Signature: (Ljava/lang/String;[B[I[B[I[B[I[I[ILjava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouis_translate
  (JNIEnv *, jobject, jstring, jbyteArray, jintArray, jbyteArray, jintArray, jbyteArray, jintArray, jintArray, jintArray, jstring, jint);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    hyphenate
 * Signature: (Ljava/lang/String;[BI[BLjava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouis_hyphenate
  (JNIEnv *, jobject, jstring, jbyteArray, jint, jbyteArray, jstring, jint);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    backTranslateString
 * Signature: (Ljava/lang/String;[B[I[B[I[BLjava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouis_backTranslateString
  (JNIEnv *, jobject, jstring, jbyteArray, jintArray, jbyteArray, jintArray, jbyteArray, jstring, jint);

/*
 * Class:     org_liblouis_LibLouis
 * Method:    backTranslate
 * Signature: (Ljava/lang/String;[B[I[B[I[I[I[I[BLjava/lang/String;I)Z
 */
JNIEXPORT jboolean JNICALL Java_org_liblouis_LibLouis_backTranslate
  (JNIEnv *, jobject, jstring, jbyteArray, jintArray, jbyteArray, jintArray, jintArray, jintArray, jintArray, jbyteArray, jstring, jint);

#ifdef __cplusplus
}
#endif
#endif
/* Header for class org_liblouis_LogCallback */

#ifndef _Included_org_liblouis_LogCallback
#define _Included_org_liblouis_LogCallback
#ifdef __cplusplus
extern "C" {
#endif
#ifdef __cplusplus
}
#endif
#endif
